#include "halType.h"
#include "halNVMHandler.h"
#include "halSPI.h"
#include "halUART.h"
#include <RTL.h>
#include "IAP.h"
#include "string.h"

#define write_Size 		4096
#define buff_Size 		1024
#define Sector_Size		32768
#define Base_Sector_No	16



#define IAP_LOCATION 0x1FFF1FF1
IAP iap_entry;


unsigned char Write_Flag =1;
//unsigned char buff[buff_Size];
//unsigned char buff1[buff_Size];

#define LAST_SECTOR_ADDR			491520
#define LAST_SECTOR_NO				29

#define CCLK						30000//(SystemFrequency/1000)
//#define START_ADDR					0x00010000

#define START_ADDR					65536
#define LAST_SECTOR_ADDR			491520


#define FLASH_FIRST_SECTOR_NO 16
#define FLASH_SECTOR_SIZE 32768
#define FLASH_START_ADDR 65536
#define FLASH_SWAP_SECTOR (29-FLASH_FIRST_SECTOR_NO)
#define FLASH_WRITE_SIZE 4096



#define ReadDByte(Address)	((unsigned char )(*(unsigned char *)Address))


#define Flash_Select()  FlashSPI_CS_Low();/** Select the Flash */
#define Flash_DeSelect() FlashSPI_CS_High ();/** DeSelect the Flash */
		  
#define WAIT_TIME os_dly_wait(5);
void Read_SPI_BYTE(int Count);

#define FLASH_OP_RETRY_MAX 30
u8 CRC7_POLY = 0x91;

u8 CycleSectorBuf[NVM_SECTOR_SIZE];

void initIAP()
{	 iap_entry =(IAP) IAP_LOCATION;
}

/***************************************************************************
Function Name : getCRC(u8PTR buf, u16 length, u8 CRC_byte)
Description   : Function to Finding CRC opf Data.


Return Type	  : Take Something Return Something.
****************************************************************************/
u8 getCRC(u8PTR buf, u16 length, u8 CRC_byte)
{
	u8 i, j, crc = CRC_byte;

	for (i = 0; i < length; i++)
	{
		crc ^= buf[i];
		for (j = 0; j < 8; j++)
		{
	  		if (crc & 1)
	    	crc ^= CRC7_POLY;
	  		crc >>= 1;
		}
	}
  	return crc;
}


#ifdef FLASHTYPE


__inline void WriteAddr32(u32 Addr)
{
	FlashSPI_SendByte(Addr>>24);
	FlashSPI_SendByte(Addr>>16);
	FlashSPI_SendByte(Addr>>8);
	FlashSPI_SendByte(Addr>>0);
}

__inline u32 GetSectorStartingAddr(u32 Addr)
{
	return (Addr/NVM_SECTOR_SIZE)*NVM_SECTOR_SIZE;
}

__inline u32 GetPageStartingAddr(u32 Addr)
{
	return (Addr/NVM_PAGE_SIZE)*NVM_PAGE_SIZE;
}

/***************************************************************************
Function Name : EraseWholeFlash()
Description   : Function to Erasing All Flash.


Return Type	  : Take Nothing Return Something.
****************************************************************************/
u8 EraseWholeFlash()
{
	int rVal=0;
	int RetryCount = 0;

	Flash_Select();		FlashSPI_SendByte(0x06);		Flash_DeSelect();
	
	Flash_Select();
	FlashSPI_SendByte(0xC7);
	Flash_DeSelect();

	Flash_Select();
	FlashSPI_SendByte (0x05);//Read Status Register

	RetryCount=0;
	while(1)
	{			
		rVal = FlashSPI_RecvByte();
		if((rVal & 0x01) == 0)
		{
			Flash_DeSelect();
			return 0;
		}

		RetryCount++;
		if(RetryCount > FLASH_OP_RETRY_MAX*1000)
		{
			Flash_DeSelect();
			Print_0("...............................Whole Flash Erase error....\r\n");
			return 3;//Sector erase error
		}
		WAIT_TIME;WAIT_TIME;WAIT_TIME;WAIT_TIME;WAIT_TIME;
	}
}
/***************************************************************************
Function Name : EraseFlashSector(int Addr)
Description   : Function to Erasing Flash Sector.


Return Type	  : Take Something Return Something.
****************************************************************************/

u8 EraseFlashSector(int Addr)
{
	int rVal=0;
	int RetryCount = 0;

	Flash_Select();		FlashSPI_SendByte(0x06);		Flash_DeSelect();//Write Enable
	
	Flash_Select();
	FlashSPI_SendByte(0x21);//Sector erase
	WriteAddr32(Addr);
	Flash_DeSelect();

	Flash_Select();
	FlashSPI_SendByte (0x05);//Read Status Register

	RetryCount=0;
	while(1)
	{			
		rVal = FlashSPI_RecvByte();
		if((rVal & 0x01) == 0)
		{
			Flash_DeSelect();
			return 0;
		}

		RetryCount++;
		if(RetryCount > FLASH_OP_RETRY_MAX)
		{
			Flash_DeSelect();
			Print_0("...............................Sector Erase error....\r\n");
			return 3;//Sector erase error
		}
		WAIT_TIME;
	}
	
}

/***************************************************************************
Function Name : ReadFromFlash(u32 StartAdd, u8PTR Buf, u32 Length)
Description   : Function to reading data from flash.
Arguments    : u32 StartAdd - Starting address 
				u8PTR Buf    - Buffer for storing data
				u32 Length   - Length of data

Return Type	  : Take Something Return Something.
****************************************************************************/
u8 ReadFromFlash(u32 StartAdd, u8PTR Buf, u32 Length)
{
	u32 EndAdd = StartAdd+Length;	

	if(StartAdd > NVM_END_ADDR)
	{
		Print_0("...............................Sector Read error: Wrong Address 1....\r\n");
		return 1;
	}
	if(StartAdd+Length > NVM_END_ADDR)
	{
		Print_0("...............................Sector Read error: Wrong Address 2....\r\n");
		return 2;
	}
	
//	Flash_DeSelect();WAIT_TIME
//	
//	Flash_Select();
//	FlashSPI_SendByte (0x05);//Read Status Register
//	rVal = FlashSPI_RecvByte();
//	Flash_DeSelect();
//	if(rVal & 0x02)
//	{
//		Print_0("...............................Sector Read error: Already in write mode....\r\n");
//		return 3;// Already in write mode.
//	}
//	if(rVal & 0x01)
//	{
//		Print_0("...............................Sector Read error: Erase process is going on....\r\n");
//		return 4;// Erase process is going on
//	}
	

	Flash_Select();	
	FlashSPI_SendByte (0x13);//Read memory
	WriteAddr32(StartAdd);
	while(StartAdd < EndAdd)
	{
		*Buf = FlashSPI_RecvByte();
		Buf++;
		StartAdd++;
	}
	Flash_DeSelect();

	return 0;
}

/***************************************************************************
Function Name : IsSectorEraseNeeded(u8PTR SectorBuf, u8PTR DataBuf, u32 Len)
Description   : Function to erasing sector data from flash.
Arguments    : u8PTR SectorBuf - Hold 0x00 
				u8PTR DataBuf  - Buffer for storing data
				u32 Length     - Length of data

Return Type	  : Take Something Return Something.
****************************************************************************/

u8 IsSectorEraseNeeded(u8PTR SectorBuf, u8PTR DataBuf, u32 Len)
{
	for(; Len; Len--, SectorBuf++, DataBuf++)
	{
		if((*SectorBuf == 0xFF) || (*DataBuf == 0x00))		continue;
		if(*SectorBuf != *DataBuf)	return TRUE;
	}

	return FALSE;
}

/***************************************************************************
Function Name : IsPageWriteNeeded(u8PTR PageBuf, u8PTR DataBuf, u32 Len)
Description   : Function to Writing page.
Arguments    :  u8PTR SectorBuf - Hold 0x00 
				u8PTR DataBuf  - Buffer for storing data
				u32 Length     - Length of data

Return Type	  : Take Something Return Something.
****************************************************************************/

u8 IsPageWriteNeeded(u8PTR PageBuf, u8PTR DataBuf, u32 Len)
{
	for(; Len; Len--, PageBuf++, DataBuf++)
	{
		if(*PageBuf != *DataBuf)	return TRUE;
	}

	return FALSE;
}

/***************************************************************************
Function Name : WriteNVMPage(u32 StartAddr, u8PTR Buf)
Description   : Function to Writing page.
Arguments    :  u32 StartAddr  - Starting address
				u8PTR Buf  - Buffer for storing data
				
Return Type	  : Take Something Return Nothing.
****************************************************************************/
void WriteNVMPage(u32 StartAddr, u8PTR Buf)
{
	u16 b=0;

	Flash_Select();		FlashSPI_SendByte(0x06);		Flash_DeSelect();//Write Enable
	Flash_Select();
	FlashSPI_SendByte(0x12);//Page program
	WriteAddr32(StartAddr);
	for(b=0; b<NVM_PAGE_SIZE; b++)
	{
		FlashSPI_SendByte(*Buf);
		Buf++;
		StartAddr++;
	}
	Flash_DeSelect();
	WAIT_TIME;
}

/***************************************************************************
Function Name : void WriteNVMSector(u32 nvmWriteAddr, u8PTR SectorBuf, u32 SectorBufFront, u8PTR SourceBuf, u32 SourceBufLen, u8 IsSectorErased)
Description   : Function to Writing NVM Sector.
				
Return Type	  : Take Something Return Nothing.
****************************************************************************/

void WriteNVMSector(u32 nvmWriteAddr, u8PTR SectorBuf, u32 SectorBufFront, u8PTR SourceBuf, u32 SourceBufLen, u8 IsSectorErased)
{
	u32 pCount = 0;
	u32 pByte = 0;
	u32 sByte = 0;
	u8 IsPageWriteNeeded = FALSE;

	for(pCount=0; pCount<NVM_SECTOR_SIZE/NVM_PAGE_SIZE; pCount++, nvmWriteAddr+=NVM_PAGE_SIZE)
	{
		for(pByte=0; pByte<NVM_PAGE_SIZE; pByte++, sByte++)
		{
			if(sByte >= SectorBufFront+SourceBufLen)	continue;
			if(sByte < SectorBufFront)	continue;
			
			if(SectorBuf[sByte] != *SourceBuf)
			{
				SectorBuf[sByte] = *SourceBuf;
				IsPageWriteNeeded = TRUE;
			}
			SourceBuf++;
		}
		if(IsPageWriteNeeded == TRUE || IsSectorErased == TRUE)
		{
			//sprintf(tmpStr, "Page write needed: %d\r\n",  nvmWriteAddr);	Print_0(tmpStr);
			WriteNVMPage(nvmWriteAddr, &SectorBuf[(pCount*NVM_PAGE_SIZE)]);
			IsPageWriteNeeded = FALSE;
		}
	}
}


/***************************************************************************
Function Name : WriteToFlash(u32 nvmWriteAddr, u8PTR SourceBuf, u32 SourceDataLen)
Description   : Function to write data on flash.
Arguments     : u32 nvmWriteAddr  - Starting address 
				u8PTR SourceBuf   - Buffer for storing data
				u32 SourceDataLen - Length of data

Return Type	  : Take Something Return Something.
****************************************************************************/
u8 WriteToFlash(u32 nvmWriteAddr, u8PTR SourceBuf, u32 SourceDataLen)
{
	u8 rVal = 0;
	u32 SectorStartAddr, SectorBufFront=0, SectorLenToCompare = 0;

	//sprintf(tmpStr, ", Writing at %d\r\n",  nvmWriteAddr);	Print_0(tmpStr);

	if(nvmWriteAddr > NVM_END_ADDR)
	{
		Print_0(".........................................WriteToFlash Error: Wrong Address 1........\r\n");
		return 1;
	}
	if(((nvmWriteAddr + SourceDataLen) > NVM_END_ADDR) || ((nvmWriteAddr + SourceDataLen) < nvmWriteAddr))
	{
		Print_0(".........................................WriteToFlash Error: Wrong Address 2........\r\n");
		return 2;
	}

//	Flash_DeSelect();WAIT_TIME
//	Flash_Select();		FlashSPI_SendByte(0x04);		Flash_DeSelect();//Write Disable
//	Flash_Select();
//	FlashSPI_SendByte (0x05);//Read Status Register
//	rVal = FlashSPI_RecvByte();
//	Flash_DeSelect();
//	if(rVal & 0x02)
//	{
//		Print_0(".........................................WriteToFlash Error: Already in write mode........\r\n");
//		return 3;// Already in write mode.
//	}
//	else if(rVal & 0x01)
//	{
//		Print_0(".........................................WriteToFlash Error: Erase process is going on........\r\n");
//		return 4;// Erase process is going on
//	}

	while(SourceDataLen)
	{
		SectorStartAddr = GetSectorStartingAddr(nvmWriteAddr);

		rVal = ReadFromFlash(SectorStartAddr, CycleSectorBuf, NVM_SECTOR_SIZE);
		if(rVal != 0)	return rVal;

		SectorBufFront = nvmWriteAddr-SectorStartAddr;
	 	SectorLenToCompare = ((NVM_SECTOR_SIZE-SectorBufFront) < SourceDataLen) ? (NVM_SECTOR_SIZE-SectorBufFront) : SourceDataLen;
		if(IsSectorEraseNeeded(&CycleSectorBuf[SectorBufFront], SourceBuf, SectorLenToCompare) == TRUE)
		{
			//sprintf(tmpStr, "Sector Erase needed: %d....\r\n",  SectorStartAddr);	Print_0(tmpStr);
			rVal = EraseFlashSector(SectorStartAddr);
			if(rVal != 0)	return rVal;
			//memset(CycleSectorBuf, 0xFF, sizeof(CycleSectorBuf));
			WriteNVMSector(SectorStartAddr, CycleSectorBuf, SectorBufFront, SourceBuf, SectorLenToCompare, TRUE);
		}
		else
			WriteNVMSector(SectorStartAddr, CycleSectorBuf, SectorBufFront, SourceBuf, SectorLenToCompare, FALSE);
		

		SourceDataLen -= SectorLenToCompare;
		nvmWriteAddr += SectorLenToCompare;
		SourceBuf += SectorLenToCompare;
	}

	//DumpMMRY();
	return 0;
}


#else

#define FLASH_WAIT_TIME 5

__inline unsigned char ReadCharFromFlash(unsigned int Addr)
{
	Addr += FLASH_START_ADDR;
	return ReadDByte(Addr);
}

/***************************************************************************
Function Name : ReadFromFlash(unsigned int Addr, unsigned char *buff, unsigned int length)
Description   : Function to read data from flash.
Arguments     : unsigned int Addr   - Starting address 
				unsigned char *buff - Buffer for storing data
				unsigned int length - Length of data

Return Type	  : Take Something Return Something.
****************************************************************************/

void ReadFromFlash(unsigned int Addr, unsigned char *buff, unsigned int length)
{
	int j;

	Addr += FLASH_START_ADDR;

	for	(j=0;j<length;j++)
	{
		//sprintf(tmpStr, "pointers Addr - Value: %d - %c\r\n",Addr, ReadDByte( (unsigned long)Addr)); Print_Log(tmpStr);
		buff[j] = ReadDByte(Addr);
		Addr++;
	}	
}

int PrepareSector(unsigned int SectorNo)
{
	unsigned int Cmd[5];
	unsigned int Rval[5];

	Cmd[0] = 50;
	Cmd[1] = SectorNo;
	Cmd[2] = SectorNo;
	Cmd[3] = CCLK;
	iap_entry (Cmd, Rval);
	os_dly_wait(FLASH_WAIT_TIME);
	if(Rval[0]!=0)
	{
		return Rval[0];
	}

	return 0;
}

/***************************************************************************
Function Name : EraseAllSectors()
Description   : Function to Erase All Sectors Data.

Return Type	  : Take Nothing Return Something.
****************************************************************************/

int EraseAllSectors()
{
	unsigned int Cmd[5];
	unsigned int Rval[5];

	Cmd[0] = 50;
	Cmd[1] = FLASH_FIRST_SECTOR_NO;
	Cmd[2] = FLASH_SWAP_SECTOR + FLASH_FIRST_SECTOR_NO;
	Cmd[3] = CCLK;
	iap_entry (Cmd, Rval);
	os_dly_wait(FLASH_WAIT_TIME);
	if(Rval[0]!=0)
	{
		//Print_0("Error in Erase ALL -1  sector....!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\r\n");
		return Rval[0];	
	}

	Cmd[0] = 52;
	Cmd[1] = FLASH_FIRST_SECTOR_NO;
	Cmd[2] = FLASH_SWAP_SECTOR + FLASH_FIRST_SECTOR_NO;
	Cmd[3] = CCLK;
	iap_entry (Cmd, Rval);
	os_dly_wait(FLASH_WAIT_TIME);
	if(Rval[0]!=0)
	{
		//Print_0("Error in Erase ALL -2  sector....!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\r\n");
		return Rval[0];	
	}
	return 0;
}

/***************************************************************************
Function Name : EraseFlashSector(unsigned int SectorNo)
Description   : Function to Erase Sector Data.

Return Type	  : Take Something Return Something.
****************************************************************************/

int EraseFlashSector(unsigned int SectorNo)
{
	unsigned int Cmd[5];
	unsigned int Rval[5];

	SectorNo += FLASH_FIRST_SECTOR_NO;
	PrepareSector(SectorNo);
		
	Cmd[0] = 52;
	Cmd[1] = SectorNo;
	Cmd[2] = SectorNo;
	Cmd[3] = CCLK;
	//Print_0("E_S");
	iap_entry (Cmd, Rval);
	os_dly_wait(FLASH_WAIT_TIME);
	//Print_0(" E_E\r\n");
	if(Rval[0]!=0)
	{
		//Print_0("Error in Erase -2  sector....!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\r\n");
		return Rval[0];	
	}
	return 0;
}

/***************************************************************************
Function Name : WriteSector(unsigned int SectorNo, unsigned int Addr, unsigned int DataLength)
Description   : Function to Writing Data on Sector.
Arguments     : unsigned int SectorNo   -  Sector No for writing.
				unsigned int Addr       -  Starting Address
				unsigned int DataLength	-  Data Length

Return Type	  : Take Something Return Something.
****************************************************************************/

int WriteSector(unsigned int SectorNo, unsigned int Addr, unsigned int DataLength)
{
	unsigned int Cmd[5];
	unsigned int Rval[5];

	SectorNo += FLASH_FIRST_SECTOR_NO;
	Addr += FLASH_START_ADDR;
	PrepareSector(SectorNo);

	Cmd[0] = 51;
	Cmd[1] = Addr;
	Cmd[2] = (unsigned int)&CycleSectorBuf;
	Cmd[3] = DataLength;
	Cmd[4] = CCLK;
	//Print_0("W_S");
	iap_entry (Cmd, Rval);
	os_dly_wait(FLASH_WAIT_TIME);
	//Print_0(" W_E\r\n");
	if(Rval[0]!=0)	
	{
		//char tmpStr[40];
		//sprintf(tmpStr, "ERR write sector: %d - %d - %d -- %d ....!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\r\n", Rval[0], SectorNo, Addr, DataLength);
		//Print_0(tmpStr);
		return Rval[0];
	}

	return 0;
}


void SwapSector(unsigned int SourceSector)
{
	unsigned int j, Count;
	unsigned int SourcePTR;
	unsigned int DestinationPTR;
	unsigned int DestinationSector = FLASH_SWAP_SECTOR;

	EraseFlashSector(FLASH_SWAP_SECTOR);

	SourcePTR = SourceSector*FLASH_SECTOR_SIZE;
	DestinationPTR = DestinationSector*FLASH_SECTOR_SIZE;
	

	for(Count=FLASH_SECTOR_SIZE; Count; Count-=FLASH_WRITE_SIZE, DestinationPTR+=FLASH_WRITE_SIZE)
	{
		for(j=0; j<FLASH_WRITE_SIZE; j++, SourcePTR++)
		{
			CycleSectorBuf[j] = ReadCharFromFlash(SourcePTR);
		}
		WriteSector(FLASH_SWAP_SECTOR, DestinationPTR, FLASH_WRITE_SIZE);
	}
}

/***************************************************************************
Function Name : WriteToFlash(unsigned int Addr, unsigned char *dataBuf, unsigned int Length)
Description   : Function to write data on flash.
Arguments     : Addr      - Starting address 
				dataBuf   - Buffer for storing data
				Length    - Length of data

Return Type	  : Take Something Return Something.
****************************************************************************/
int WriteToFlash(unsigned int Addr, unsigned char *dataBuf, unsigned int Length)
{
	unsigned int  flashPTR, bufPTR, fp, bp, i, j, swapPTR;
	unsigned int s, SectorEraseNeeded = FALSE;
	u8 fStartSector, fEndSector;

	fStartSector = Addr / FLASH_SECTOR_SIZE;
	fEndSector = (Addr+Length-1) / FLASH_SECTOR_SIZE;
	flashPTR = Addr;
	bufPTR=0;

	for(s=fStartSector; s<=fEndSector; s++)
	{
		SectorEraseNeeded = FALSE;
		for(fp=0,bp=0; (fp+flashPTR)<((s+1)*FLASH_SECTOR_SIZE) && (bp+bufPTR)<Length; fp++, bp++)
		{
			if(SectorEraseNeeded == TRUE)	continue;
		 	if((ReadCharFromFlash(flashPTR+fp) != dataBuf[bufPTR+bp]))// && (ReadCharFromFlash(flashPTR+fp) != 0xFF))
				SectorEraseNeeded = TRUE;
		}
		if(s>fStartSector)	Addr = 0;
		
		//sprintf(tmpStr, "..............SectorEraseNeeded: %d\r\n", SectorEraseNeeded); Print_Log(tmpStr);
		if(SectorEraseNeeded == TRUE)
		{
			SwapSector(s);			//Print_0("Sector Swap Done....\r\n");
			EraseFlashSector(s);  	//Print_0("Sector Erase Done....\r\n");

			swapPTR = FLASH_SECTOR_SIZE*FLASH_SWAP_SECTOR;
			for(i=0; i<FLASH_SECTOR_SIZE; i+=FLASH_WRITE_SIZE)
			{
				memset(CycleSectorBuf, 0xFF, FLASH_WRITE_SIZE);
				for(j=0; j<FLASH_WRITE_SIZE; j++, swapPTR++)
				{
					if(((i+j) < Addr) || (bufPTR>=Length))
					{
						CycleSectorBuf[j] = ReadCharFromFlash(swapPTR);
					}
					else
					{
						CycleSectorBuf[j] = dataBuf[bufPTR];
						bufPTR++;
					}
					//CycleSectorBuf[j] = (((i+j)<Addr) || (bufPTR>=Length)) ? ReadCharFromFlash(swapPTR) : dataBuf[bufPTR];
				}
				//Print_0("Write ");
				WriteSector(s, (s*FLASH_SECTOR_SIZE)+i, FLASH_WRITE_SIZE);
				//Print_0("Done\r\n");
			}
		}
		flashPTR += fp;
	}
	return 0;
}

#endif


